"""
Functions related to timeseries management.

This file can also be imported as a module and contains the following
functions:

    * format_to_timetable - converts wind speed to power output of a wind turbine
    * synchronise - convert solar irradiance to the power output of a PV installation
    * resample_timeseries - resamples a timeseries to a given resolution
    * split_in_daily_profiles - splits a timeseries in a set of daily profiles
    * split_timeseries_in_chunks - splits a timeseries in chunks of data
    * timestep_start - determines a timestep start for a given resolution
    * now - determines a the current time (floored to a given resolution)
    * datetime_to_microepoch - converts a datetime to microepoch
    * microepoch_to_datetime_index - converts microepoch to datetime
    * microepoch_to_local_datetime - converts microepoch to a local datetime
    
"""

from datetime import datetime, timedelta
from typing import Any

import numpy as np

# Import all useful libraries
import pandas as pd
from dateutil.tz import tzlocal
from pytz import timezone

from .types import (
    timezone_input,
    timezone_output,
    dict_or_none,
    bool_or_none,
    str_or_none,
    datetime_or_none,
)


# Helpers
def __is_timezone(var: Any) -> bool:
    class_name = str(var.__class__)
    return class_name.startswith("<class 'pytz.tzfile.") or class_name.startswith(
        "<class 'dateutil.tz.tz."
    )


def __get_local_timezone() -> timezone_output:
    return tzlocal()


# Dealing with timeseries data
def format_to_timetable(
    data: pd.DataFrame,
    time_column: str = "t",
    column_dict: dict_or_none = None,
    inplace: bool = False,
    infer_datetime: bool = True,
    utc_index: bool_or_none = None,
):
    """Formats a pandas dataframe into a timetable

    Parameters
    ----------
    data : pd.DataFrame
        Dataframe to convert to timetable
    time_column : str
        Name of the column to use as time index
        (default: "t")
    column_dict : dict
        Dictionary to use for column renaming, if needed
        (default: None)
    inplace : Boolean
        Whether or not to operate the formatting in place
        (default: False)
    infer_datetime : Boolean
        Whether or not to infer the datetime format from the first element.
        See pandas.to_datetime() documentation for more details
        (default: True (for speed considerations in API context) )
    utc_index : Boolean
        If True, ensures that the timezone is defined to UTC on the index (time_column)
        (default: None)

    Raises
    ------
    None

    Returns
    -------
    data : pd.DataFrame
        DataFrame indexed by time (timetable)
        (None - if inplace is True)
    """

    if inplace is True:
        d = data
    else:
        d = data.copy()

    if len(d) > 0:
        if column_dict is not None:
            d.rename(columns=column_dict, inplace=True)

        if time_column != "t":
            d.rename(columns={time_column: "t"}, inplace=True)

        if isinstance(d["t"][0], str):
            d["t"] = pd.to_datetime(
                d["t"], infer_datetime_format=infer_datetime, utc=utc_index
            )

        d.set_index("t", inplace=True)
        d.sort_index(ascending=True, inplace=True)

    if inplace is True:
        return None
    else:
        return d


def synchronise(
    df1: pd.DataFrame, df2: pd.DataFrame, base: str = "first", fill: str_or_none = None
):
    """Formats a pandas dataframe into a timetable

    Parameters
    ----------
    df1 : pd.DataFrame
        First DataFrame to merge
    df2 : pd.DataFrame
        First DataFrame to merge
    base : str
        Time base for the merged DataFrame
        (options: "first","second","intersection","union")
        (default: "first")
    fill : str
        Filling method, if relevant
        (options: "backfill")
        (default: None)

    Raises
    ------
    None

    Returns
    -------
    data : pd.DataFrame
        DataFrame indexed by time (timetable)
    """

    # Some more hints here: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#resampling

    if base == "first":
        df_m = df1.merge(df2, how="left", right_index=True, left_index=True)

    elif base == "second":
        df_m = df1.merge(df2, how="right", right_index=True, left_index=True)

    elif base == "intersection":
        df_m = df1.merge(df2, how="intersection", right_index=True, left_index=True)

    elif base == "union":
        df_m = df1.merge(df2, how="outer", right_index=True, left_index=True)

    else:
        raise Exception(f"This case is not supported : {base}")

    if fill is not None:
        if fill == "backfill":
            df_m = df_m.fillna("backfill", axis="index")
        else:
            raise Exception(f"This fill is not supported yet: {fill}")

    return df_m


def resample_timeseries(
    data: pd.DataFrame,
    resolution: str = "raw",
    function: str = "mean",
    tz: timezone_input = None,
    **kwargs,
) -> pd.DataFrame:
    """Resamples a timeseries at a given frequency

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame to resample
    resolution : str
        Resolution to use for the daily profile
        (options: "raw","hour","day")
        (default: "raw")
    function : str
        Function to use for the resampling
        (options: "mean","min","max","sum")
        (default: "mean")
    tz : str, pytz timezone, or None
        Time-zone for the daily profile
        If None is specified, or if the index is timezone-naive, this is skipped
        (default: None)

    Raises
    ------
    None

    Returns
    -------
    T_split : pd.DataFrame
        DataFrame of containing the split of the daily profiles, indexed by date
    """

    need_to_reset_timezone = False
    target_tz = __backwards_compatibility_tz(tz=tz, legacy_tz=kwargs.get("timezone"))

    if resolution == "raw" or data is None or len(data) == 0:
        None

    else:
        if resolution == "hour":
            resampler = data.resample("H")
        elif resolution == "day":
            if target_tz is None:
                resampler = data.resample("D")
            else:
                # Averaging per day is made in local timezone
                data.index = data.index.tz_convert(target_tz)
                resampler = data.resample("D")
                need_to_reset_timezone = True

        else:
            raise Exception(f"Illegal value of data resolution: {resolution}")

        if function == "mean":
            data = resampler.mean()
        elif function == "sum":
            data = resampler.sum()
        elif function == "min":
            data = resampler.min()
        elif function == "max":
            data = resampler.max()
        else:
            raise Exception(f"Illegal value of function: {function}")

    if need_to_reset_timezone is True:
        data.index = data.index.tz_convert("UTC")

    return data


def split_in_daily_profiles(
    data: pd.DataFrame,
    column: str,
    resolution: str = "hour",
    tz: str_or_none = None,
    aggregate: str = "mean",
    **kwargs,
) -> pd.DataFrame:
    """Formats a pandas dataframe into a timetable

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame to split in daily profiles
    column : str
        Name of the column to split in daily profiles
    resolution : str
        Resolution to use for the daily profile
        (options: "hour","15min","10min","5min")
        (default: "hour")
    tz : str, pytz timezone, or None
        Time-zone for the daily profile
        If None is specified, or if the index is timezone-naive, this is skipped
        (default: None)
    aggregate : str
        How to aggregate points if downsampling
        (options: "mean", "sum")
        (default: "mean")

    Raises
    ------
    None

    Returns
    -------
    T_split : pd.DataFrame
        DataFrame of containing the split of the daily profiles, indexed by date
    """
    target_tz = __backwards_compatibility_tz(tz=tz, legacy_tz=kwargs.get("time_zone"))

    data_copy = data[[column]].copy()
    data_copy["date"] = data_copy.index.date

    if target_tz is not None:
        data_copy.index = data_copy.index.tz_convert(target_tz)

    T_split = (
        pd.DataFrame({"date": data_copy.index.date})
        .drop_duplicates(subset=["date"])
        .set_index("date")
    )

    if resolution == "hour":
        data_copy["timeOfDay"] = data_copy.index.hour
        range2use = range(0, 24, 1)
        col_name = lambda x: "H" + str(x)

    elif resolution == "30min":
        data_copy["timeOfDay"] = data_copy.index.hour * 60 + data_copy.index.minute
        range2use = range(0, 24 * 60, 30)
        col_name = lambda x: "M15_" + str(int(x / 30))

    elif resolution == "15min":
        data_copy["timeOfDay"] = data_copy.index.hour * 60 + data_copy.index.minute
        range2use = range(0, 24 * 60, 15)
        col_name = lambda x: "M15_" + str(int(x / 15))

    elif resolution == "10min":
        data_copy["timeOfDay"] = data_copy.index.hour * 60 + data_copy.index.minute
        range2use = range(0, 24 * 60, 10)
        col_name = lambda x: "M10_" + str(int(x / 10))

    elif resolution == "5min":
        data_copy["timeOfDay"] = data_copy.index.hour * 60 + data_copy.index.minute
        range2use = range(0, 24 * 60, 5)
        col_name = lambda x: "M5_" + str(int(x / 5))

    else:
        raise Exception("Illegal resolution: " + resolution)

    for i in range2use:
        T_i = data_copy[[column, "date"]].loc[data_copy["timeOfDay"] == i]
        if aggregate == "sum":
            T_i = T_i.groupby(["date"]).sum()
        elif aggregate == "mean":
            T_i = T_i.groupby(["date"]).mean()
        else:
            raise Exception(f"Illegal value of aggregate ({aggregate})")
        T_i = T_i.rename(columns={column: col_name(i)})
        T_split = T_split.merge(T_i, how="left", right_index=True, left_index=True)

    return T_split


def split_data_loading_range(
    start: datetime, end: datetime, step: timedelta = timedelta(weeks=24)
):
    date_range = []
    max_date = lambda x: np.min([end, x + step])
    time_1 = start
    time_2 = start
    while time_2 < end:
        time_1 = time_2
        time_2 = max_date(time_1)
        date_range.append([time_1, time_2])

    return date_range


def split_timeseries_in_chunks(timeseries: pd.Series, length: int) -> pd.DataFrame:
    """
    Formats a pandas series into chunks (e.g. for machine learning usages)

    Parameters
    ----------
    timeseries : pd.Series
        Timeseries to split in chunks
    length : int
        Length of the chunks

    Raises
    ------
    None

    Returns
    -------
    chunks_df : pd.DataFrame
        DataFrame of chunks (columns names [x0, x1, ..., xN] where N=length-1)
    """
    chunks_df = pd.DataFrame()
    for i in range(0, length):
        chunks_df[f"x{i}"] = timeseries.shift(-i)
    return chunks_df


def timestep_start(
    step: str, t: datetime_or_none = None, tz: timezone_input = __get_local_timezone()
) -> datetime:

    if t is None:
        t = datetime.now(tz)

    if step == "raw":
        t_start = t
    elif step in ["second", "1s"]:
        t_start = t.replace(microsecond=0)
    elif step == "15s":
        sec_start = int(t.second / 15) * 15
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step == "30s":
        sec_start = int(t.second / 30) * 30
        t_start = t.replace(microsecond=0, second=sec_start)
    elif step in ["minute", "1min"]:
        t_start = t.replace(microsecond=0, second=0)
    elif step == "5min":
        min_start = int(t.minute / 5) * 5
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "15min":
        min_start = int(t.minute / 15) * 15
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step == "30min":
        min_start = int(t.minute / 30) * 30
        t_start = t.replace(microsecond=0, second=0, minute=min_start)
    elif step in ["hour", "1h"]:
        t_start = t.replace(microsecond=0, second=0, minute=0)
    elif step in ["day", "1d"]:
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0)
    elif step == "month":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1)
    elif step == "year":
        t_start = t.replace(microsecond=0, second=0, minute=0, hour=0, day=1, month=1)
    else:
        raise Exception(f"Unknown step: {step}")

    return t_start


def now(step: str = "raw", tz: timezone_input = __get_local_timezone()) -> datetime:
    return timestep_start(step, t=None, tz=tz)


# Functions dealing with microepoch data
def datetime_to_microepoch(datetimes2use: datetime) -> int:
    return int(datetimes2use.timestamp() * 1e6)


def microepoch_to_datetime_index(microepoch: pd.Series) -> pd.DatetimeIndex:
    t = pd.DatetimeIndex(np.round(microepoch).astype("datetime64[us]"))
    return t.tz_localize("UTC")


def microepoch_to_local_datetime(microepoch) -> pd.Series:
    t = pd.Datetime(np.round(microepoch).astype("datetime64[us]"))
    return t.tz_localize(__get_local_timezone())


# For backwards compatibility
def __backwards_compatibility_tz(
    tz: timezone_input = None, legacy_tz: timezone_input = None
) -> timezone_output:
    if legacy_tz is None:
        tz_out = tz
    else:
        if tz is None:
            tz_out = legacy_tz
        else:
            raise Exception(
                """Parameter name used for timezone is only used for backwards compatibility -
                please update to 'tz' to support future standard"""
            )

    if tz_out is None:
        return None
    elif isinstance(tz_out, str):
        return timezone(tz_out)
    elif __is_timezone(tz_out):
        return tz_out
    else:
        raise Exception(
            "Invalid format for input 'time_zone' (must be either str or pytz timezone)"
        )



from seedftw.exceptions import MovedToPlotneat


def group_legend_by_name(*args, **kwargs):
    raise MovedToPlotneat()


def minimalistic_show(*args, **kwargs):
    raise MovedToPlotneat()


def clean_legend(*args, **kwargs):
    raise MovedToPlotneat()



from datetime import datetime, timedelta
from typing import Callable

import pandas as pd
from seedftw.base.timeseries import split_data_loading_range


# Loading data from the API for 1 parameter
def split_api_calls(
    func: Callable,
    start: datetime,
    end: datetime,
    max_step: timedelta,
    margin_last: timedelta = timedelta(seconds=1),
) -> pd.DataFrame:
    """
    Function to split loading of data in a smart manner
    """

    # Create the steps for loading
    date_steps_range = split_data_loading_range(start=start, end=end, step=max_step)

    data = None
    for start_j, end_j in date_steps_range:
        data_j = func(start_j, end_j - margin_last)

        if data is None:
            data = data_j
        else:
            data = data.append(data_j)

    return data



"""
Functions related to modelling of energy networks.

This file can also be imported as a module and contains the following
functions:

    * split_import_export - split timeseries of import and export


"""

from typing import Tuple
import numpy as np
import pandas as pd


def split_import_export(net_import: pd.Series) -> Tuple[pd.Series, pd.Series]:
    """Models the power production curve of a standard wind turbine.
    The model is based upon (Twidell, 2006), pp. 306-307.

    Parameters
    ----------
    net_import : np.array or Series
        Values of the net import (i.e. >0 is import, <0 is export)

    Raises
    ------
    None

    Returns
    -------
    imports : np.array, Series or pd.DataFrame
        Timeseries of imports
    exports : np.array, Series or pd.DataFrame
        Timeseries of exports
    """

    if net_import is None:
        imports = None
        exports = None
    elif all(v is None for v in net_import):
        imports = net_import
        exports = net_import
    else:
        imports = np.fmax(net_import, [0])
        exports = np.fmax(-net_import, [0])

    return imports, exports



# -*- coding: utf-8 -*-
"""
Functions related to modelling of energy assets.

This file can also be imported as a module and contains the following
functions:

    [None] Refer to package "senasopt" on PyPi

Note: the previous functions have been deprecated.

"""

from seedftw.exceptions import MovedToSenasopt


def __moved_to_senasopt(new_function):
    raise MovedToSenasopt(new_function)


def wind_turbine_generator(*args):
    __moved_to_senasopt(new_function="wind.wind_turbine")


def solar_pv_generator(*args):
    __moved_to_senasopt(new_function="solar.pv_module")


def battery_optimal_controller(*args):
    __moved_to_senasopt(new_function="battery.battery_optimal_controller")



import errno
import json
import os
import warnings
from typing import Dict

from . import _version
from . import base
from . import energy
from . import environment
from . import exceptions

__version__ = _version.get_versions()["version"]


USER_LEVEL_CONFIGURATION_FILE = os.path.expanduser("~/.seedftw/config.json")

__env_var_configuration_loaded = "SEEDFTW_TOKEN_DK_DMI_METOBSAPI_V2"


def __base_configuration() -> Dict:
    return {
        "dmi_metObsAPIv1": "",
        "dmi_metObsAPIv2": "",
    }


def load_configuration(force_reload: bool = False) -> None:
    # Avoiding to reload the configuration if already loaded
    if (force_reload is False) and (
        os.environ.get(__env_var_configuration_loaded) == "1"
    ):
        return
    elif os.path.exists(USER_LEVEL_CONFIGURATION_FILE) is False:
        raise Exception(
            f"The configuration file does not exist ({USER_LEVEL_CONFIGURATION_FILE})"
        )
    else:
        os.environ[__env_var_configuration_loaded] = "1"

    from seedftw.environment.denmark import (
        set_api_token as set_environment_dk_api_token,
    )

    with open(USER_LEVEL_CONFIGURATION_FILE, "r") as fp:
        api_keys_in_file = json.load(fp)

    if api_keys_in_file.get("dmi_metObsAPIv1") is not None:
        set_environment_dk_api_token(
            dmi_metObsAPI=api_keys_in_file.get("dmi_metObsAPIv1")
        )

    if api_keys_in_file.get("dmi_metObsAPIv2") is not None:
        set_environment_dk_api_token(
            dmi_metObsAPIv2=api_keys_in_file.get("dmi_metObsAPIv2")
        )


def create_configuration_file(
    overwrite: bool = False, open_for_edit: bool = False
) -> None:

    if (overwrite is False) and os.path.exists(USER_LEVEL_CONFIGURATION_FILE):
        warnings.warn(
            f"The configuration file already exists ({USER_LEVEL_CONFIGURATION_FILE})"
        )

    else:
        base_config = __base_configuration()

        if not os.path.exists(os.path.dirname(USER_LEVEL_CONFIGURATION_FILE)):
            try:
                os.makedirs(os.path.dirname(USER_LEVEL_CONFIGURATION_FILE))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        # Writing to sample.json
        with open(USER_LEVEL_CONFIGURATION_FILE, "w") as outfile:
            outfile.write(json.dumps(base_config, indent=4))

    if open_for_edit:
        os.startfile(USER_LEVEL_CONFIGURATION_FILE, "open")



class MissingApiKeyError(Exception):
    """Raised when an API key has not been loaded.

    Attributes:
        message -- explanation of why the specific transition is not allowed
        environment_variable -- enviroment variable that should contain the key
    """

    def __init__(self, message, environment_variable, create_at_url):
        print(
            "You are missing an API loaded in the environment variable "
            + environment_variable
            + ".\nIf you do not have one, create one on: "
            + create_at_url
            + "\n\n"
        )
        self.message = message
        self.environment_variable = environment_variable
        self.create_at_url = create_at_url


class MovedToPlotneat(Exception):
    def __init__(self):
        super().__init__(
            """
        This code was moved to plotneat, which can be downloaded here:
        https://gitlab.com/Pierre_VF/plotneat
        """
        )


class MovedToSenasopt(Exception):
    def __init__(self, new_function):
        super().__init__(
            """This function is deprecated and has been removed.

            To access a similar functionality, refer to the package "senasopt" on PyPi.

            The corresponding function in the package is: senasopt.{}
            (you might however have to do slight aaptation of your code at the interface)

            """.format(
                new_function
            )
        )



"""
Functions related to data loading in Denmark.

API specification is found on:
    https://confluence.govcloud.dk/display/FDAPI/Meteorological+Observation

functions:

    * get_weather_stations - loads the list of all Danish weather stations
    * get_weather_station_data - loads historical data for a given station
    * get_closest_weather_station - identifies the closest weather station
    * get_data_for_coordinates - loads historical data from the closest station
        to a set of coordinates
    * set_api_token - loads the API token

"""

import os
import warnings
from datetime import datetime, timedelta
from functools import lru_cache
from typing import List, Union, Dict

import numpy as np
import pandas as pd
from pytz import utc
from requests import Session, Response
from seedftw.base.timeseries import (
    format_to_timetable,
    datetime_to_microepoch,
    microepoch_to_datetime_index,
    synchronise,
)
from seedftw.base.timeseries import resample_timeseries
from seedftw.base.tricks import split_api_calls
from seedftw.environment.geography import distance_between_coordinates
from seedftw.exceptions import MissingApiKeyError

__dk_timezone = "Europe/Copenhagen"
__env_var_api_token_metobsapi = "SEEDFTW_TOKEN_DK_DMI_METOBSAPI"
__env_var_api_token_metobsapiv2 = "SEEDFTW_TOKEN_DK_DMI_METOBSAPI_V2"
__latest_version = "v2"

__session = Session()


def __warn_deprecation(version="v1"):
    warnings.warn(
        "Usage of this version ({}) will soon be deprecated".format(version),
        PendingDeprecationWarning,
    )


def __unsupported_version(version):
    raise Exception("Unsupported version ({})".format(version))


def set_api_token(
    dmi_metObsAPIv2: Union[str, None] = None, dmi_metObsAPI: Union[str, None] = None
) -> None:
    """Loads the API token for the metObsAPI from the Danish weather stations.

    If you do not have one, go to https://confluence.govcloud.dk/pages/viewpage.action?pageId=26476690 to create one.

    Parameters
    ----------
    dmi_metObsAPIv2 : str
            API token for the metObsAPI (v2).

    dmi_metObsAPI : str
            API token for the metObsAPI (v1 - deprecated).

    Raises
    ------
    None

    Returns
    -------
    None
    """
    if dmi_metObsAPIv2 is not None:
        os.environ[__env_var_api_token_metobsapiv2] = dmi_metObsAPIv2
        # headers = {"Authorization" : "X-Gravitee-Api-Key: " + str(dmi_metObsAPIv2)}
        headers = {"X-Gravitee-Api-Key": str(dmi_metObsAPIv2)}
        __session.headers.update(headers)

    if dmi_metObsAPI is not None:
        __warn_deprecation(version="v1")
        os.environ[__env_var_api_token_metobsapi] = dmi_metObsAPI


def _get_api_key(api: str = "MetObsAPIv2") -> str:

    if api == "MetObsAPI":
        __warn_deprecation(version="v1")
        dmi_api_key = os.getenv(__env_var_api_token_metobsapi)
        if dmi_api_key is None:
            e = MissingApiKeyError(
                message="Missing API key for DMI service.",
                environment_variable=__env_var_api_token_metobsapi,
                create_at_url="https://confluence.govcloud.dk/display/FDAPI",
            )
            raise e
    elif api == "MetObsAPIv2":
        dmi_api_key = os.getenv(__env_var_api_token_metobsapiv2)
        if dmi_api_key is None:
            e = MissingApiKeyError(
                message="Missing API key for DMI service.",
                environment_variable=__env_var_api_token_metobsapi,
                create_at_url="https://confluence.govcloud.dk/display/FDAPI",
            )
            raise e
    else:
        raise Exception("Illegal api input ({})".format(api))

    return dmi_api_key


def __format_data_table(data):
    return format_to_timetable(
        data, time_column="Minutes5UTC", column_dict=None, utc_index=True
    )


# Loading data from DMI's API
def __dmi_api_call(
    address: str, params: Union[Dict, None] = None, version: str = __latest_version
) -> Response:

    if version == "v1":
        __warn_deprecation(version="v1")
        if params is not None:
            raise Exception("Usage of params is not supported in v1")
        api_key_part = "api-key=" + _get_api_key(api="MetObsAPI")
        query_url = "https://dmigw.govcloud.dk/metObs/v1/" + address
        if "?" in query_url:
            binding_sym = "&"
        else:
            binding_sym = "?"
        # Read data in Json format
        url_2_call = query_url + binding_sym + api_key_part
        result = pd.read_json(url_2_call)

    elif version == "v2":
        query_url = "https://dmigw.govcloud.dk/v2/metObs/collections/" + address
        result = __session.get(query_url, params=params).json()

    else:
        __unsupported_version(version)

    return result


# Loading data from the API for 1 parameter
def __sub_load_parameter_timeseries(
    station: Union[str, int],
    parameter: str,
    start: datetime,
    end: datetime,
    version: str = __latest_version,
) -> pd.DataFrame:
    """
    Documentation of the details of the data is available here:
    >>  https://confluence.govcloud.dk/display/FDAPI/Meteorological+Observation
    """

    if isinstance(station, str):
        station_string = station
    else:
        station_string = str(station).zfill(5)

    parameter2id = {
        "ambient_temperature": "temp_dry",
        "wind_speed": "wind_speed",
        "wind_direction": "wind_dir",
        "relative_humidity": "humidity",
        "global_solar_radiation": "radia_glob",
    }

    # Adding a query size limiter to approx. 6 months
    limit2use = 30000

    if version == "v1":
        start_microepoch = datetime_to_microepoch(start)
        end_microepoch = datetime_to_microepoch(end)
        query_url = (
            "observation?from={}&to={}&stationId={}&parameterId={}&limit={}".format(
                str(start_microepoch),
                str(end_microepoch),
                station_string,
                parameter2id[parameter],
                str(limit2use),
            )
        )
        params = None
    elif version == "v2":

        def __dateformat(t):
            return t.astimezone(utc).isoformat(timespec="seconds", sep="T")

        query_url = "observation/items"
        params = {
            "datetime": "{}/{}".format(__dateformat(start), __dateformat(end)),
            "stationId": station_string,
            "parameterId": parameter2id[parameter],
            "limit": str(limit2use),
        }
    else:
        __unsupported_version(version)

    if version == "v1":
        dmi_raw_data = __dmi_api_call(query_url, params=params, version="v1")
        dmi_raw_data["t"] = microepoch_to_datetime_index(dmi_raw_data["timeObserved"])
    elif version == "v2":
        dmi_raw_data_json = __dmi_api_call(query_url, params=params, version="v2")
        dmi_raw_data = pd.DataFrame(dmi_raw_data_json["features"])["properties"].apply(
            pd.Series
        )
        dmi_raw_data["t"] = dmi_raw_data["observed"]
    else:
        __unsupported_version(version)

    data = format_to_timetable(
        dmi_raw_data[["t", "value"]],
        time_column="t",
        column_dict={"value": parameter},
        utc_index=True,
    )

    if version == "v2":
        data.sort_index(axis=0, inplace=True, ascending=True)

    return data


def __load_parameter_timeseries(
    station, parameter, start, end, resolution, version=__latest_version
):

    # Split calls
    max_step = timedelta(weeks=24)
    margin_last = timedelta(seconds=1)

    def func(t0, t1):
        return __sub_load_parameter_timeseries(
            station,
            parameter,
            t0,
            t1,
            version=version,
        )

    data = split_api_calls(func, start, end, max_step, margin_last)

    # Resample
    data = resample_timeseries(
        data, resolution=resolution, function="mean", tz=__dk_timezone
    )

    return data


@lru_cache(maxsize=2)
def __raw_load_weather_stations(version: str = __latest_version) -> pd.DataFrame:
    if version == "v1":
        call2use = "station?country=DNK"
        df = __dmi_api_call(call2use, params=None, version="v1")
        stations = df["location"].apply(pd.Series)
        focus_pars = ["name", "type", "stationId"]
        stations[focus_pars] = df[focus_pars]

    elif version == "v2":
        call2use = "station/items"
        res = __dmi_api_call(call2use, params=None, version="v2")
        stations = pd.DataFrame(res["features"])["properties"].apply(pd.Series)
        stations["type"] = pd.DataFrame(res["features"])["type"].apply(pd.Series)
        stations[["longitude", "latitude"]] = (
            pd.DataFrame(res["features"])["geometry"]
            .apply(pd.Series)["coordinates"]
            .tolist()
        )
        stations["validTo"] = pd.to_datetime(
            stations["validTo"], infer_datetime_format=True, utc=True
        )
        stations["validFrom"] = pd.to_datetime(
            stations["validFrom"], infer_datetime_format=True, utc=True
        )
        stations["operationTo"] = pd.to_datetime(
            stations["operationTo"], infer_datetime_format=True, utc=True
        )
        stations["operationFrom"] = pd.to_datetime(
            stations["operationFrom"], infer_datetime_format=True, utc=True
        )

    else:
        __unsupported_version(version)

    return stations


def get_weather_stations(
    only_available: bool = False,
    version: str = __latest_version,
    valid_since: Union[None, datetime] = None,
) -> pd.DataFrame:
    """Loads the list of the Danish weather stations

    Parameters
    ----------
    version : version of the DMI API ("v2" or "v1")

    Raises
    ------
    None

    Returns
    -------
    stations : pandas.DataFrame(height,latitude,longitude,name,type,stationId)
        Details of the weather stations
    """
    stations = __raw_load_weather_stations(version=version)

    if version == "v1":
        if only_available is True:
            stations = stations.loc[stations["type"] == "Synop", :]
        if valid_since is not None:
            raise Exception("Usage of valid_since is not supported in version v1")

    elif version == "v2":
        if only_available is True:
            # Filtering all stations not currently available
            k_valid = [np.isnat(x) for x in stations["validTo"].values]
            stations = stations[k_valid]

        if valid_since is not None:
            # Filtering all stations with data start after the allowed minimum
            stations = stations[stations["validFrom"] <= valid_since]
    else:
        __unsupported_version(version)

    return stations


def get_weather_station_data(
    station: Union[int, str] = 6031,
    start: datetime = None,
    end: datetime = None,
    resolution: str = "hour",
    parameters: List[str] = ["ambient_temperature"],
    version: str = __latest_version,
) -> pd.DataFrame:
    """Loads historical data for a given weather station

    Parameters
    ----------
    station : int or str
        stationId of the station for which data is to be loaded
    start : datetime
        Time for the start of the historical period
    end: datetime
        Time for the end of the historical period
    resolution: str
        Resolution of the data to load (raw,hour,day)
    parameters : list(str)
        List of parameters to be loaded
        (Options: "ambient_temperature","wind_speed","wind_direction",
         "relative_humidity","global_solar_radiation")
    version : version of the DMI API ("v2" or "v1")

    Raises
    ------
    None

    Returns
    -------
    data : pandas.DataFrame(t,ambient_temperature)
        Timetable of the ambient temperature [degrees C]


    Documentation of the details of the data is available here:
    >>  https://confluence.govcloud.dk/pages/viewpage.action?pageId=26476616

    """

    if start is None:
        start = datetime.now() - timedelta(days=3)
    if end is None:
        end = datetime.now()

    data = None

    for par_i in parameters:
        data_i = __load_parameter_timeseries(
            station, par_i, start, end, resolution, version=version
        )
        if data is None:
            data = data_i
        else:
            data = synchronise(data, data_i, "union")

    return data


def get_closest_weather_station(
    latitude: float,
    longitude: float,
    version: str = __latest_version,
    country: str = "dk",
) -> pd.Series:
    """Identifies the closest weather station for a set geographical coordinates

    Parameters
    ----------
    latitude : float
        Latitude [degrees]
    longitude : float
        Longitude [degrees]
    version : version of the DMI API ("v2" or "v1")

    Raises
    ------
    None

    Returns
    -------
    closest_station : pandas.Series(height,latitude,longitude,name,type,stationId)
        Details of the closest location
    """

    # Get all DMI locations
    selected_stations = get_weather_stations(only_available=True, version=version)

    if version == "v2":
        selected_stations = selected_stations[
            [("temp_dry" in s_i) for s_i in selected_stations["parameterId"]]
        ]

        low_country = country.lower()
        if low_country in ["dk", "dnk", "denmark"]:
            selected_stations = selected_stations[selected_stations["country"] == "DNK"]
        else:
            raise Exception("Unsupported land ({})".format(country))

    distance_to_stations = selected_stations.apply(
        lambda row: distance_between_coordinates(
            [latitude, longitude], [row["latitude"], row["longitude"]]
        ),
        axis=1,
    )

    closest_index = selected_stations.index[np.argmin(distance_to_stations)]
    closest_station = selected_stations.loc[closest_index, :]

    return closest_station


def get_data_for_coordinates(
    latitude: str = 57.046707,
    longitude: str = 9.935932,
    start: datetime = (datetime.now() - timedelta(days=3)),
    end: datetime = datetime.now(),
    resolution: str = "hour",
    parameters: List[str] = ["ambient_temperature"],
    version: str = __latest_version,
) -> pd.DataFrame:
    """Loads historical data from the closest weather station to a set of coordinates

    Parameters
    ----------
    latitude : float
        Latitude [degrees]
    longitude : float
        Longitude [degrees]
    start : datetime
        Time for the start of the historical period
    end: datetime
        Time for the end of the historical period
    resolution: str
        Resolution of the data to load (raw,hour,day)
    parameters : list(str)
        List of parameters to be loaded
        (Options: "ambient_temperature","wind_speed","wind_direction",
         "relative_humidity","global_solar_radiation")
    version : version of the DMI API ("v2" or "v1")

    Raises
    ------
    None

    Returns
    -------
    data : pandas.DataFrame(t,ambient_temperature)
        Timetable of the ambient temperature [degrees C]
    """

    closest_station = get_closest_weather_station(latitude, longitude, version=version)
    data = get_weather_station_data(
        station=closest_station["stationId"],
        start=start,
        end=end,
        resolution=resolution,
        parameters=parameters,
        version=version,
    )
    return data



from seedftw.exceptions import MovedToPlotneat
import pytest


def test_import():
    import seedftw

    seedftw.load_configuration()
    assert isinstance(seedftw.__version__, str)


def test_moved():
    from seedftw.visual import figure, plot
    from seedftw.exceptions import MovedToSenasopt
    from seedftw.energy.energy_asset import (
        wind_turbine_generator,
        solar_pv_generator,
        battery_optimal_controller,
    )

    # Plotneat
    with pytest.raises(MovedToPlotneat):
        figure.group_legend_by_name()
    with pytest.raises(MovedToPlotneat):
        figure.clean_legend()
    with pytest.raises(MovedToPlotneat):
        figure.minimalistic_show()

    # Senasopt
    with pytest.raises(MovedToPlotneat):
        plot.duration_curve()
    with pytest.raises(MovedToSenasopt):
        wind_turbine_generator()
    with pytest.raises(MovedToSenasopt):
        solar_pv_generator()
    with pytest.raises(MovedToSenasopt):
        battery_optimal_controller()



# -*- coding: utf-8 -*-
"""
Functions related to solar radiatiom modelling.

This file can also be imported as a module and contains the following
functions:

    * hour_angle - Computes the solar hour angle
    * declination - Computes the solar declination 
    * day_length - Computes the day length
    * solar_azimuth_angle - Computes the solar azimuth angle
    * zenith_angle - Computes the solar zenith angle 
    * solar_altitude - Computes the solar altitude angle
    * air_mass_ratio - Computes the air-mass ratio 

Work to do:
    * beam_surface_angle - Implement the function
    * implement calls for above functions with lists and dataframe columns

References:
    [1] J. Twidell & T. Weir, Renewable Energy Resources, 2nd edition, 
    Taylor and Francis, 2006

"""

from datetime import datetime

import numpy as np
from pytz import timezone

from .geography import cosd, sind
from seedftw.base.timeseries import timestep_start


# Sine-related functions in degrees
def __tand(x):
    return np.tan(x * (2 * np.pi / 360))


def __secd(x):
    return 1 / cosd(x)


def __arccosd(x):
    return 360 / (2 * np.pi) * np.arccos(x)


def __deg2pm180(x):
    return np.mod(x + 180, 360) - 180


# Solar constant ([1] p. 86) - in W/m2
__solar_constant = 1367

# Earth declination ([1] p. 90) - in degrees
__earth_declination = 23.45


def hour_angle(t: datetime) -> float:
    """Computes the solar hour

    Parameters
    ----------
    t : datetime
        Time for which to compute the hour angle

    Returns
    -------
    w : float
        Solar hour angle [degrees]
    """

    t_utc = t.astimezone(timezone("UTC"))
    w = 15 * (t_utc.hour - 12)
    return w


def declination(t: datetime) -> float:
    """Computes the solar declination

    Parameters
    ----------
    t : datetime
        Time for which to compute the declination

    Returns
    -------
    d : float
        Solar declination [degrees]
    """
    # From [1] p. 91, adjusted for radians
    t_utc = t.astimezone(timezone("UTC"))
    t_year_start = timestep_start("year", t=t)
    n = (t_utc - t_year_start).days + 1
    d = __earth_declination * np.sin(2 * np.pi * (284 + n) / 365)
    return d


def day_length(t: datetime, latitude: float) -> float:
    """Computes the day length

    Parameters
    ----------
    t : datetime
        Time for which to compute the day length
    latitude : float
        Geographical latitude [degrees]

    Returns
    -------
    daylight_hours : float
        Day length [h]
    """
    # From [1] p. 91, adjusted for radians
    d = declination(t)
    temp = -__tand(latitude) * __tand(d)
    if np.abs(temp) <= 1:
        daylight_hours = 2 / 15 * __arccosd(temp)
    else:
        if temp > 0:
            daylight_hours = 0
        else:
            daylight_hours = 24

    return daylight_hours


def solar_azimuth_angle(t: datetime, longitude: float) -> float:
    """Computes the solar azimuth angle

    Parameters
    ----------
    t : datetime
        Time for which to compute the azimuth angle
    longitude : datetime
        Longitude for which to compute the azimuth angle [degrees]

    Returns
    -------
    azi : float
        Solar azimuth [degrees]
    """
    # From [1] p.96
    w = hour_angle(t)
    gamma_s = w - longitude
    return __deg2pm180(gamma_s)


def zenith_angle(t: datetime, latitude: float):
    """Computes the solar zenith angle

    Parameters
    ----------
    t : datetime
        Time for which to compute the zenith angle
    latitude : datetime
        Geographical latitude for which to compute the zenith angle [degrees]

    Returns
    -------
    theta_z : float
        Zenith angle [degrees]
    """
    # From [1] p.96
    d = declination(t)
    phi = latitude
    w = hour_angle(t)
    theta_z = __arccosd(sind(phi) * sind(d) + cosd(phi) * cosd(w) * cosd(d))
    return theta_z


def solar_altitude(t: datetime, latitude: float) -> float:
    """Computes the solar altitude angle

    Parameters
    ----------
    t : datetime
        Time for which to compute the solar altitude angle
    latitude : datetime
        Geographical latitude for which to compute the solar altitude angle [degrees]

    Returns
    -------
    alt : float
        Solar altitude angle [degrees]
    """
    # From [1] p.96
    theta_z = zenith_angle(t, latitude)
    return 90 - theta_z


def air_mass_ratio(t: datetime, latitude: float) -> float:
    """Computes the air-mass ratio

    Parameters
    ----------
    t : datetime
        Time for which to compute the air-mass ratio
    latitude : datetime
        Geographical latitude for which to compute the air-mass ratio [degrees]

    Returns
    -------
    AM : float
        Air-mass ratio [/]
        Note: This ratio is truncated below 1 and above 100 to ensure meaningfulness
    """
    # From [1] p.98
    theta_z = zenith_angle(t, latitude)
    AM = __secd(theta_z)
    if AM <= 1 or AM > 100:
        AM = np.nan
    return AM


def beam_surface_angle(t: datetime, latitude: float) -> float:
    # From [1] p.95
    raise Exception("To be implemented")



# -*- coding: utf-8 -*-
"""
Functions related to modelling of geography-related aspects.

This file can also be imported as a module and contains the following
functions:

    * distance_between_coordinates - computes distance between coordinates


References:
    [1] "Reference points and distance computations" (PDF), Code of Federal Regulations (Annual Edition). 
        Title 47: Telecommunication. 73 (208). October 1, 2016.
        URL: https://www.govinfo.gov/content/pkg/CFR-2016-title47-vol4/pdf/CFR-2016-title47-vol4-sec73-208.pdf

"""
import warnings
from typing import List

import numpy as np


# Sine and cosine in degrees
def cosd(x):
    return np.cos(x * (2 * np.pi / 360))


def sind(x):
    return np.sin(x * (2 * np.pi / 360))


def distance_between_coordinates(
    coordinates_1: List[float], coordinates_2: List[float]
) -> float:
    """Computes the distance between two sets of geographical coordinates

    Parameters
    ----------
    coordinates_1 : list(2)
        Coordinates in format [latitude,longitude] [degrees]
    coordinates_2 : list(2)
        Coordinates in format [latitude,longitude] [degrees]

    Raises
    ------
    None

    Returns
    -------
    distance : float
        Distance between the coordinates [km]
    """

    latitude_1 = coordinates_1[0]
    longitude_1 = coordinates_1[1]
    latitude_2 = coordinates_2[0]
    longitude_2 = coordinates_2[1]

    # With formulates from reference [1]
    latitude_M = (latitude_1 + latitude_2) / 2
    K1 = 111.13209 - 0.56605 * cosd(2 * latitude_M) + 0.00120 * cosd(4 * latitude_M)
    K2 = (
        111.41513 * cosd(latitude_M)
        - 0.09455 * cosd(3 * latitude_M)
        + 0.00012 * cosd(5 * latitude_M)
    )

    distance = np.sqrt(
        (K1 * (latitude_1 - latitude_2)) ** 2 + (K2 * (longitude_1 - longitude_2)) ** 2
    )

    if distance >= 475:
        warnings.warn(
            "WARNING: The distance estimation is not accurate at this distance ("
            + str()
            + " km > 475 km)"
        )
        distance = np.NAN

    return distance



from datetime import datetime, tzinfo
from typing import Union

timezone_input = Union[str, tzinfo, None]
timezone_output = Union[tzinfo, None]
dict_or_none = Union[dict, None]
bool_or_none = Union[bool, None]
str_or_none = Union[str, None]
datetime_or_none = Union[datetime, None]



# -*- coding: utf-8 -*-
"""
Functions related to data loading in Great Britain.

Sources:
- https://carbonintensity.org.uk/

"""

from datetime import timedelta

# For caching
from functools import lru_cache

import numpy as np
import pandas as pd
import requests
from dateutil.parser import parse
from seedftw.base.timeseries import (
    format_to_timetable,
    resample_timeseries,
    timestep_start,
    now,
)
from seedftw.base.tricks import split_api_calls

__uk_timezone = "Europe/London"
__api_url = "https://api.carbonintensity.org.uk/"

__api_region_ids = {
    "north scotland": 1,
    "south scotland": 2,
    "north west england": 3,
    "north east england": 4,
    "yorkshire": 5,
    "north wales": 6,
    "south wales": 7,
    "west midlands": 8,
    "east midlands": 9,
    "east england": 10,
    "south west england": 11,
    "south england": 12,
    "london": 13,
    "south east england": 14,
    "england": 15,
    "scotland": 16,
    "wales": 17,
}
# Loading data from CarbonIntensity.org.uk
@lru_cache(maxsize=10)
def __api_call(endpoint):
    r = requests.get(__api_url + endpoint)

    data = r.json()

    if data is None:
        raise Exception(
            "No data returned by the API (on endpoint: {})".format(endpoint)
        )

    if "error" in data.keys():
        status = data["error"]["code"]
    else:
        status = 200

    if status == 200:
        None
    elif status.startswith("400"):
        raise Exception("Bad request (on endpoint: {})".format(endpoint))
    elif status.startswith("500"):
        raise Exception("Internal server error (on endpoint: {})".format(endpoint))
    else:
        raise Exception(
            "Unknown return status={} (on endpoint: {})".format(status, endpoint)
        )

    return data


def __convert_string_to_datetime(string):
    return parse(string)


def get_emission_factors(out="dict"):
    endpoint = "intensity/factors"
    r = __api_call(endpoint)
    data = r["data"][0]

    low_format = out.lower()

    if low_format == "dict":
        return data
    elif low_format == "dataframe":
        return pd.DataFrame(
            {"technology": data.keys(), "emission_intensity": data.values()}
        )
    else:
        raise Exception("Unknown format: " + out)


def __get_regional_api_endpoint(area, start, end):
    if isinstance(area, str):
        if area.lower() in ["england", "wales", "scotland"]:
            endpoint = "regional/intensity/{}/{}/regionid/{}".format(
                start.isoformat(), end.isoformat(), __api_region_ids[area.lower()]
            )
        else:
            endpoint = "regional/intensity/{}/{}/postcode/{}".format(
                start.isoformat(), end.isoformat(), area
            )
    else:
        endpoint = "regional/intensity/{}/{}/regionid/{}".format(
            start.isoformat(), end.isoformat(), str(area)
        )
    return endpoint


def __unpack_region_data(data):
    data_copy = data["data"].copy().apply(pd.Series)
    return data_copy


def __load_historical_intensity(start, end, area):
    if area is None:
        endpoint = "intensity/{}/{}".format(start.isoformat(), end.isoformat())
    else:
        endpoint = __get_regional_api_endpoint(area, start, end)

    r = __api_call(endpoint)
    raw_data = pd.DataFrame(r["data"])

    if "regionid" in raw_data.keys():
        raw_data = __unpack_region_data(raw_data)

    if len(raw_data) == 0:
        raise Exception("No data available")

    raw_data["start"] = pd.to_datetime(raw_data["from"])

    temp_i = raw_data["intensity"].apply(pd.Series)
    raw_data[temp_i.keys()] = temp_i

    for col_i in ["forecast", "actual"]:
        if col_i not in raw_data.keys():
            raw_data[col_i] = np.NaN

    data = format_to_timetable(
        raw_data[["start", "forecast", "actual"]],
        time_column="start",
        column_dict=None,
        utc_index=True,
    )
    return data


def __get_historical_intensity(
    start=None,
    end=None,
    resolution="raw",
    area=None,
):
    if start is None:
        start = timestep_start("30min") - timedelta(days=2)
    elif end is None:
        end = timestep_start("30min")

    max_step = timedelta(days=13)
    fun2use = lambda x, y: __load_historical_intensity(x, y, area)
    data = split_api_calls(
        fun2use,
        start,
        end,
        max_step,
        margin_last=timedelta(seconds=1),
    )

    data = resample_timeseries(
        data, resolution=resolution, function="mean", tz=__uk_timezone
    )
    return data


def __split_demand_percentages(raw_data):
    y = None
    for x_i in raw_data["generationmix"]:
        x_i = pd.DataFrame(x_i).set_index("fuel").transpose()
        if y is None:
            y = x_i
        else:
            y = pd.concat([y, x_i])

    y.index = raw_data.index

    for col_i in y:
        raw_data[col_i] = y[col_i]

    raw_data.pop("generationmix")


def __load_historical_generation(start, end, area=None):

    if area is None:
        endpoint = "generation/{}/{}".format(start.isoformat(), end.isoformat())
    else:
        endpoint = __get_regional_api_endpoint(area, start, end)

    r = __api_call(endpoint)
    raw_data = pd.DataFrame(r["data"])

    if "regionid" in raw_data.keys():
        raw_data = __unpack_region_data(raw_data)
        raw_data.pop("intensity")

    raw_data["start"] = pd.to_datetime(raw_data["from"])

    __split_demand_percentages(raw_data)
    raw_data.pop("from")
    raw_data.pop("to")

    data = format_to_timetable(
        raw_data, time_column="start", column_dict=None, utc_index=True
    )
    return data


def get_historical_generation(
    start=None,
    end=None,
    resolution="raw",
    area=None,
):
    if start is None:
        start = timestep_start("30min") - timedelta(days=2)
    elif end is None:
        end = timestep_start("30min")

    max_step = timedelta(days=13)
    fun2use = lambda x, y: __load_historical_generation(x, y, area)
    data = split_api_calls(
        fun2use,
        start,
        end,
        max_step,
        margin_last=timedelta(seconds=1),
    )

    data = resample_timeseries(
        data, resolution=resolution, function="mean", tz=__uk_timezone
    )
    return data


def electricity_average_co2_intensity(
    area=None,
    start=None,
    end=None,
    resolution="raw",
):
    if start is None:
        start = now() - timedelta(hours=168)
    elif end is None:
        end = now()
    return __get_historical_intensity(start, end, resolution, area=area)


def electricity_average_co2_intensity_forecast(
    area=None,
    start=None,
    end=None,
    resolution="raw",
):
    if start is None:
        start = now()
    elif end is None:
        end = now() + timedelta(hours=24)
    return __get_historical_intensity(start, end, resolution, area=area)



"""
Functions related to data loading in Denmark.

"""

import urllib.parse
from datetime import datetime, timedelta

# For caching
from functools import lru_cache

# Import all useful libraries
import pandas as pd
from seedftw.base.timeseries import format_to_timetable, now


# Loading data from Energinet
@lru_cache(maxsize=10)
def __dk_energinet_dataservice_call(sql_query, output_format="Dataframe"):

    # Determine the URL to use for the call
    url = (
        "https://www.energidataservice.dk/proxy/api/datastore_search_sql?sql="
        + urllib.parse.quote(sql_query)
    )

    # Read data in Json format
    api_data = pd.read_json(url)

    if any(api_data.success) == False:
        print("Failure in the API call")

    if output_format == "Dataframe":
        api_data = pd.DataFrame.from_records(api_data["result"].records)

    return api_data


def __format_time(t_in):
    if isinstance(t_in, datetime):
        t = t_in
    elif isinstance(t_in, str):
        t = pd.to_datetime(t_in)
    else:
        raise Exception("Illegal type for time (only supporting string and datetime")
    return t.strftime("%Y-%m-%dT%H:%M:%SZ")


def __format_start_end(start, end):
    start = __format_time(start)
    end = __format_time(end)
    return start, end


def __format_data_table_co2_per_area(data):
    column_dict = {
        "CO2Emission": "co2_intensity",
        "PriceArea": "price_area",
    }
    return format_to_timetable(
        data, time_column="Minutes5UTC", column_dict=column_dict, utc_index=True
    )


def __format_data_table_co2_per_distribution(data):
    column_dict = {
        "Co2PerkWh": "co2_intensity",
        "PriceArea": "price_area",
        "Co2PerkWh_200": "co2_intensity_200pc",
    }
    return format_to_timetable(
        data, time_column="HourUTC", column_dict=column_dict, utc_index=True
    )


def electricity_average_co2_intensity(
    area="",
    start=None,
    end=None,
):
    if start is None:
        start = now() - timedelta(hours=168)
    if end is None:
        end = now()

    start, end = __format_start_end(start, end)

    if area == "DK1":
        condition2use = """WHERE "PriceArea" = 'DK1' AND"""

    elif area == "DK2":
        condition2use = """WHERE "PriceArea" = 'DK2' AND"""

    else:
        condition2use = """WHERE"""

    query2use = f"""SELECT "Minutes5UTC", "PriceArea", "CO2Emission" FROM "co2emis" {condition2use} "Minutes5UTC" >= '{start}' AND "Minutes5UTC" <= '{end}' ORDER BY "Minutes5UTC" ASC"""

    co2_data = __dk_energinet_dataservice_call(
        sql_query=query2use, output_format="Dataframe"
    )
    return __format_data_table_co2_per_area(co2_data)


def electricity_average_co2_intensity_forecast(
    area="",
    start=None,
    end=None,
):

    if start is None:
        start = now()
    if end is None:
        end = now() + timedelta(hours=24)
    start, end = __format_start_end(start, end)

    if area == "DK1":
        condition2use = """WHERE "PriceArea" = 'DK1' AND"""

    elif area == "DK2":
        condition2use = """WHERE "PriceArea" = 'DK2' AND"""

    else:
        condition2use = """WHERE"""

    query2use = f"""SELECT "Minutes5UTC", "PriceArea", "CO2Emission" FROM "co2emisprog" {condition2use} "Minutes5UTC" >= '{start}' AND "Minutes5UTC" <= '{end}' ORDER BY "Minutes5UTC" ASC"""

    co2_data = __dk_energinet_dataservice_call(
        sql_query=query2use, output_format="Dataframe"
    )
    return __format_data_table_co2_per_area(co2_data)


def electricity_distribution_average_co2_intensity(
    start=None,
    end=None,
):
    if start is None:
        start = now() - timedelta(days=365)
    if end is None:
        end = now()

    start, end = __format_start_end(start, end)

    query2use = f"""SELECT "HourUTC", "Co2PerkWh", "Co2PerkWh_200", "PriceArea" FROM "declarationemissionhour" WHERE "HourUTC">='{start}' AND "HourUTC"<='{end}' ORDER BY "HourUTC" ASC"""

    co2_data = __dk_energinet_dataservice_call(
        sql_query=query2use, output_format="Dataframe"
    )
    return __format_data_table_co2_per_distribution(co2_data)


def electricity_balance(
    area="",
    start=None,
    end=None,
):
    # Doc available here: https://www.energidataservice.dk/tso-electricity/electricitybalance
    if start is None:
        start = now() - timedelta(days=15)
    if end is None:
        end = now()

    start, end = __format_start_end(start, end)

    if area == "DK1":
        condition2use = """ "PriceArea" = 'DK1' AND"""

    elif area == "DK2":
        condition2use = """ "PriceArea" = 'DK2' AND"""

    else:
        condition2use = ""

    query2use = (
        """SELECT "HourUTC", "GrossCon", "ElectricBoilerCon", "NetCon","""
        + """ "LocalPowerProd", "ExchangeGreatBelt","OffshoreWindPower","""
        + """ "CentralProd","ExchangeNordicCountries",	"ExchangeContinent","""
        + """ "PriceArea",	"OnshoreWindPower",	"SolarPowerProd" """
        + """FROM "electricitybalance" """
        + f"""WHERE{condition2use} "HourUTC">='{start}' AND "HourUTC"<='{end}' """
        + ' ORDER BY "HourUTC" ASC'
    )

    dict2use = {
        "GrossCon": "gross_demand",
        "ElectricBoilerCon": "electric_boiler_demand",
        "NetCon": "net_demand",
        "LocalPowerProd": "local_power_generation",
        "ExchangeGreatBelt": "exchange_great_belt",
        "OffshoreWindPower": "offshore_wind_generation",
        "CentralProd": "central_generation",
        "ExchangeNordicCountries": "exchange_nordics",
        "ExchangeContinent": "exchange_continent",
        "OnshoreWindPower": "onshore_wind_generation",
        "SolarPowerProd": "solar_generation",
        "PriceArea": "price_area",
    }

    data = __dk_energinet_dataservice_call(
        sql_query=query2use, output_format="Dataframe"
    )
    return format_to_timetable(
        data, time_column="HourUTC", column_dict=dict2use, utc_index=True
    )


def electricity_production_and_exchange(
    area="",
    start=None,
    end=None,
):
    # Doc available here: https://www.energidataservice.dk/tso-electricity/electricityprodex5minrealtime
    if start is None:
        start = now() - timedelta(days=2)
    if end is None:
        end = now()

    start, end = __format_start_end(start, end)

    if area == "DK1":
        condition2use = """ "PriceArea" = 'DK1' AND"""

    elif area == "DK2":
        condition2use = """ "PriceArea" = 'DK2' AND"""

    else:
        condition2use = ""

    condition2use = """WHERE{} "Minutes5UTC">='{}' AND "Minutes5UTC"<='{}' """.format(
        condition2use, start, end
    )

    query2use = (
        """SELECT "Minutes5UTC", "PriceArea", "ProductionLt100MW", "ProductionGe100MW","""
        + """ "OffshoreWindPower", "OnshoreWindPower", "SolarPower","""
        + """ "ExchangeGreatBelt", "ExchangeGermany","ExchangeNetherlands", "ExchangeNorway","""
        + """ "ExchangeSweden",	"BornholmSE4" """
        + """FROM "electricityprodex5minrealtime" """
        + condition2use
        + ' ORDER BY "Minutes5UTC" ASC'
    )

    dict2use = {
        "PriceArea": "price_area",
        "ProductionLt100MW": "generation_under_100MW",
        "ProductionGe100MW": "generation_over_100MW",
        "OffshoreWindPower": "offshore_wind_generation",
        "OnshoreWindPower": "onshore_wind_generation",
        "SolarPower": "solar_generation",
        "ExchangeGreatBelt": "import_from_great_belt",
        "ExchangeGermany": "import_from_germany",
        "ExchangeNetherlands": "import_from_netherlands",
        "ExchangeNorway": "import_from_norway",
        "ExchangeSweden": "import_from_sweden",
        "BornholmSE4": "import_from_sweden_to_bornholm",
    }

    data = __dk_energinet_dataservice_call(
        sql_query=query2use, output_format="Dataframe"
    )
    return format_to_timetable(
        data, time_column="Minutes5UTC", column_dict=dict2use, utc_index=True
    )


def electricity_spot_price(
    area="",
    start=None,
    end=None,
):
    if start is None:
        start = now() - timedelta(days=15)
    if end is None:
        end = now()

    start, end = __format_start_end(start, end)

    if area == "DK1":
        condition2use = """ "PriceArea" = 'DK1' AND"""

    elif area == "DK2":
        condition2use = """ "PriceArea" = 'DK2' AND"""

    else:
        condition2use = ""

    query2use = f"""SELECT "HourUTC", "SpotPriceEUR", "SpotPriceDKK", "PriceArea" FROM "elspotprices" WHERE{condition2use} "HourUTC">='{start}' AND "HourUTC"<='{end}'  ORDER BY "HourUTC" ASC"""

    dict2use = {
        "SpotPriceEUR": "spot_price_eur",
        "SpotPriceDKK": "spot_price_dkk",
        "PriceArea": "price_area",
    }

    data = __dk_energinet_dataservice_call(
        sql_query=query2use, output_format="Dataframe"
    )
    return format_to_timetable(
        data, time_column="HourUTC", column_dict=dict2use, utc_index=True
    )
